/*
 * keytab.c - This file is responsible for maintaining the data structures used
 *            by the daemon. This basically encompases the key list, and the
 *            mode tree. Each mode contains a list of actions defined within it.
 */ 

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <linux/kd.h>
#include <sys/ioctl.h>
#include "keyd.h"

char *initialModeName = "initial";
char *mixerDeviceName = "/dev/mixer";
char *cdromDeviceName = "/dev/cdrom";

struct keymode {
	char *name;
	int numkeys;  /* number of keys specified in this mode */
	struct keymode *parent;
	struct keymode *child;
	struct keymode *next;
	action_t onEntry, onExit;
	action_t downActions[MAX_KEYS];
	action_t upActions[MAX_KEYS];
};

keydrc_t keydrc = NULL;
keymode_t curmode = NULL;
keymode_t _parsemode = NULL;
keydrc_t _parserc = NULL;
void _killModeRefs();


keydrc_t makeKeydrc()
{
	keydrc_t rc = s_zmalloc(sizeof( struct keydrc_struct ));

	rc->mixerdev = strdup(mixerDeviceName);
	rc->cdromdev = strdup(cdromDeviceName);
	rc->rootmode = s_zmalloc(sizeof( struct keymode ));
	rc->rootmode->name = strdup(initialModeName);
	rc->initialmode = rc->rootmode;
	return rc;
}

void deleteModeTree(keymode_t mode)
{
	int i;
	keymode_t child;
	
	child = mode->child;
	while(child){
		deleteModeTree(child);
		child = child->next;
	}
	for( i=0; i<MAX_KEYS; i++ ){
		if( mode->downActions[i] ) free(mode->downActions[i]);
		if( mode->upActions[i] && mode->upActions[i] != mode->downActions[i] ) 
			free(mode->upActions[i]);
	}
	s_free( mode->onEntry );
	s_free( mode->onExit );
	free( mode->name );
	free( mode );
}

void deleteConfig(keydrc_t rc)
{
	int i;
	
	deleteModeTree(rc->rootmode);
	for(i=0;i<rc->numkeys;i++) 
		free(rc->keytab[i]);
	free(rc->mixerdev);
	free(rc->cdromdev);
	if(rc->startup) free(rc->startup);
	free(rc);
}

keydrc_t loadConfig(char *fname)
{
	FILE *f;
	keydrc_t rc;

	f = fopen(fname, "ro");
	if(!f){
		log(LOG_ERR, "Unable to open config file: %s (%s)",fname,strerror(errno));
		return NULL;
	}
	rc = makeKeydrc();
	_parsemode = rc->rootmode;
	_parserc = rc;
	yyrestart(f);
#ifdef DEBUGPARSER
	yydebug = 1;
#endif
	if( yyparse(rc) != 0 ){
		deleteConfig(rc);
		_killModeRefs();
		fclose(f);
		return NULL;
	}
#ifdef DEBUG
	printModes(rc);
#endif
	resolveRefs(rc);
	fclose(f);
	return rc;
}


//************** Mode List ********************

struct moderef {
	char *name;
	keymode_t from;
	mode_action_t action;
	struct moderef *next;
} *_moderefs;

keymode_t findMode(keymode_t from, char *name)
{
	/* FIXME: Change this to do breadth-first search */ 

	keymode_t mode, test;
	if( !strcasecmp(from->name, name) ) return from;
	
	mode = from->child;
	while(mode){
		test = findMode(mode, name);
		if(test) return test;
		mode = mode->next;
	}
	return NULL;
}

void _killModeRefs()
{
	struct moderef *tmp;
	
	while(_moderefs){
		tmp=_moderefs->next;
		free(_moderefs);
		_moderefs = tmp;
	}
}

void addModeRef( mode_action_t action, char *name )
{
	struct moderef *ref;
	
	ref = s_malloc(sizeof(struct moderef));
	ref->name = name;
	ref->action = action;
	ref->from = _parsemode;
	ref->next = _moderefs;
	_moderefs = ref;
}

void resolveRef( keydrc_t keyd, struct moderef *ref )
{
	keymode_t mode;
	mode = findMode( ref->from, ref->name );
	if(!mode){
		mode = ref->from;
		while(mode && strcasecmp(mode->name, ref->name)) mode = mode->parent;
		if(!mode){
			mode = findMode( keyd->rootmode, ref->name );
			if(!mode) {
				log(LOG_ERR, "Unable to resolve reference to mode: %s", ref->name );
				exit(2);
			}
		}
	}
	ref->action->mode = mode;
}

void resolveRefs( keydrc_t keyd )
{
	struct moderef *ref, *tmp;
	
	ref = _moderefs;
	while(ref){
		resolveRef(keyd, ref);
		tmp = ref->next;
		free(ref);
		ref = tmp;
	}
	_moderefs = NULL;
}



keymode_t addMode( char *name )
{
	keymode_t mode;
	keymode_t parent = _parsemode;
	
	/* FIXME: Check if mode already exists, and raise error */
	
	mode = s_malloc( sizeof(struct keymode) );
	memset(mode, 0, sizeof(struct keymode));
	mode->name = name;
	mode->parent = parent;
	mode->next = mode->parent->child;
	mode->parent->child = mode;
	_parserc->nummodes++;
	return mode;
}

void openModeBlock( keymode_t mode )
{
	if(mode) _parsemode = mode;
}

void closeModeBlock( )
{
	if(_parsemode->parent) _parsemode = _parsemode->parent;
}

void setEntryAction(action_t action)
{
	_parsemode->onEntry = action;
}

void setExitAction(action_t action)
{
	_parsemode->onExit = action;
}

void setStartupAction( action_t action)
{
	if( _parserc->startup ) {
		log( LOG_WARNING, "Multiple startup actions specified, ignoring" );
	}else
		_parserc->startup = action;
}

void addAction( keymode_t mode, int key, int up, 
                action_t action )
{
	action_t *actions = (up ? mode->upActions : mode->downActions);

	if(actions[key] == NULL )
		mode->numkeys++;
	else
		log(LOG_WARNING, "Warning: Overriding action for keycode %d (%d)\n",
			_parserc->keytab[key]->key, _parserc->keytab[key]->mod );
	actions[key] = action;
}


int addKey( keyent_t key, action_t action )
{
	int i;
	keydrc_t keyd = _parserc;
	
	for( i=0; i<keyd->numkeys; i++){
		if(keyd->keytab[i]->key == key->key && keyd->keytab[i]->mod == (key->mod & 0x1FF))
			break;
	}
	if( i == MAX_KEYS ){
		fprintf(stderr,"MAX_KEYS exceeded! (maximum 256)\n");
		return FALSE;
	}
	keyd->keytab[i] = key;
	key->orig = -1;
	if(i == keyd->numkeys) keyd->numkeys++;
	if(key->mod&MOD_UP) addAction( _parsemode, i, 1, action);
	if(key->mod&MOD_DOWN || ((key->mod&UPDOWN_MASK) == 0))
		addAction( _parsemode, i, 0, action );
	key->mod &= 0x1FF;
	return TRUE;
}


/********************* Kernel loading functions *************************/
void loadKey( int key, int mod, int val )
{
	struct kbentry kb;
	
	kb.kb_table = (char)(mod&0xFF);
	kb.kb_index = (char)(key&0xFF);
	kb.kb_value = val;
	if( ioctl(consolefd, KDSKBENT, &kb) == -1){
		log(LOG_ERR, "KDSKBENT failed! (%s)",strerror(errno));
	}
}
int readKey( int key, int mod )
{
	struct kbentry kb;
	kb.kb_table = (char)(mod&0xFF);
	kb.kb_index = (char)(key&0xFF);
	if( ioctl(consolefd, KDGKBENT, &kb) == -1) {
		log(LOG_ERR, "KDGKBENT failed! (%s)",strerror(errno));
	}
	return kb.kb_value;
}


void loadKeyTable( )
{
	int i;
  
	for( i=0; i<keydrc->numkeys; i++ ){
		keydrc->keytab[i]->orig = readKey( keydrc->keytab[i]->key, 
		                                   keydrc->keytab[i]->mod );
#ifdef DEBUG
		printf("Loading key: %d mod %04X\n",
				keydrc->keytab[i]->key, keydrc->keytab[i]->mod );
#endif
		loadKey( keydrc->keytab[i]->key, keydrc->keytab[i]->mod, K(KT_HOTKEY,i) );
	}
}

void restoreKeyTable( )
{
	int i;
	
	for( i=0; i<keydrc->numkeys; i++ ){
		if(keydrc->keytab[i]->orig != -1)
			loadKey( keydrc->keytab[i]->key, keydrc->keytab[i]->mod, 
			         keydrc->keytab[i]->orig);
		keydrc->keytab[i]->orig = -1;
	}
}

/******************* Runtime Operations ******************/

action_t findAction( int keynum, int up )
{
	keymode_t mode = curmode;
	if(up){
		while(mode->upActions[keynum] == NULL && mode->parent)
			mode = mode->parent;
		return mode->upActions[keynum];
	} else {
		while(mode->downActions[keynum] == NULL && mode->parent)
			mode = mode->parent;
		return mode->downActions[keynum];
	}	
}

void execKey( int keynum, int up )
{
	action_t action;
	
	action = findAction( keynum, up );
	if(action) execAction( action );
}

void enterMode(keymode_t mode)
{
	if(mode && mode != curmode){
#ifdef DEBUG
		log(LOG_INFO, "Entering mode: %s",mode->name );
#endif DEBUG
		if(curmode->onExit) execAction(curmode->onExit);
		if(mode->onEntry) execAction(mode->onEntry);
		curmode = mode;
	}
}

void installConfig( keydrc_t rc )
{
	keydrc_t old = NULL;
	
	if(!rc) return;
	
	if( keydrc ){
		restoreKeyTable();
		old = keydrc;
	}

	keydrc = rc;
	curmode = keydrc->initialmode;
	if(keydrc->startup) execAction(keydrc->startup);
	if(curmode->onEntry) execAction(curmode->onEntry);
	loadKeyTable();

	if(old) deleteConfig(old);
}	

// ************************ Debugging Functions ***************************

void _printModes( keydrc_t keyd, keymode_t from, int level )
{
	keymode_t mode;
	
	fprintf(stderr,"%*c%s (%d keys)\n",level*2,' ',from->name, from->numkeys );
	mode = from->child;
	while(mode){
		_printModes( keyd, mode, level+1 );
		mode = mode->next;
	}
}

void printModes( keydrc_t keyd )
{
	fprintf(stderr,"Mode tree:\n");
	_printModes(keyd, keyd->rootmode, 1);
}
	